class GameException < Exception
end

class GameController < ApplicationController
  # These will essentially be ignored, because their effect will be
  # overridden after loading
  NUM_ROUNDS     = 2
  NUM_CATEGORIES = 5
  NUM_QUESTIONS  = 5

  NUM_PLAYERS  = 4
  DEFAULT_PLAYERS = ['AIChE', 'SAE', '&Tau;&Beta;&Pi;', 'ASES']
  
  REQ_CHANGE_ROUND = ?r
  REQ_GUESS        = ?g
  REQ_QUESTION     = ?q
  REQ_ANSWER       = ?a
  REQ_DONE         = ?d

  ICON_NEG = 0
  ICON_POS = 1
  ICONS = [
    (0...1).collect { |i| "neg#{'%02d'%i}.png" },
    (0...1).collect { |i| "pos#{'%02d'%i}.png" }
  ]

  MAX_NUM_GUESSES = 2
  
#  SAMPLE = 'quiz.txt'
  SAMPLE  = 'FINAL/finals.txt'
  SAMPLE1 = 'FINAL/finals_2008.txt'
  SAMPLE2 = 'FINAL/finals_2009.txt'

  prepend_before_filter :validate_session, :except => [ :index, :setup, :clear ]

  ############################################################

  # Simply redirect to setup
  def index
  end

  ############################################################

  # Change the round and display the intro
  def round
    if params[:id]
      @round = params[:id].to_i - 1
      @round = [[@round,0].max, @board.num_rounds-1].min
      session[:round] = @round
    end
  end

  ############################################################

  # Clear's the done flag on all questions in this round
  def reset_round
    @board[@round].clear
    redirect_to :action => :board
  end

  ############################################################
  
  # Remove the loaded questions and return to setup
  def remove_board
    reset_session
    redirect_to :action => :setup
  end

  ############################################################
  
  # List all the questions and answers so that they may be printed
  # (all needed work is in the template)
  def list
  end

  ############################################################
  
  # Manually update the scores
  def correct_scores
    if request.post?
      begin
        scores = params[:scores]
        raise GameException, 'Scores unspecified' if !scores
        raise GameException, 'Scores incorrectly specified' if (!scores.kind_of? Hash or scores.size != NUM_PLAYERS)
        scores = (0...NUM_PLAYERS).collect { |i| scores[i.to_s].to_i }
        @players.size.times { |i| @players[i].score = scores[i] }
      rescue GameException => e
        flash[:notice] = e.message
      end
    end
  end

  ############################################################
  
  # Draw the board
  def board
    # Update the round if posted
    rq, id = parse_req

    case rq
    # Change round number and display intro screen
    when REQ_CHANGE_ROUND
      board_change_round(id)
    # Display question or answer
    when REQ_QUESTION, REQ_ANSWER, REQ_GUESS
      board_questions_and_answers(rq, id)
    # When the game is over, show the results
    when REQ_DONE
      board_done    
    end

    board_setup_for_rendering
  end

  ############################################################
  
  # Setup the session
  def setup
    @players = (0...NUM_PLAYERS).collect { |i| Player.new(DEFAULT_PLAYERS[i]) }

    # Load the posted questions
    if request.post?
      begin
        if params[:id] == 'sample'
          # Automatically load the sample
          @image_path = default_image_path
          @file = File.open(SAMPLE, 'r')
        elsif params[:id] == 'sample1'
          # Automatically load the sample
          @image_path = default_image_path
          @file = File.open(SAMPLE1, 'r')
        elsif params[:id] == 'sample2'
          # Automatically load the sample
          @image_path = default_image_path
          @file = File.open(SAMPLE2, 'r')
        else
          # Get the image path
          @image_path = params[:image_path]
          @image_path = default_image_path if !@image_path or @image_path.strip.size == 0
          @image_path = @image_path.strip
          
          # Ensure that a file was uploaded
          @file = params[:file]
          raise IOError, "Please choose a quiz file to upload." if @file.size == 0

          # Extract the player names
          players = params[:players]
          if players and players.kind_of? Hash
            for i in 0...NUM_PLAYERS
              pname = players[i.to_s]
              if pname and pname.size > 0
                @players[i].name = pname
              end
            end
          end
        end

        # Load the board and set up the session variables
        set_image_path(@image_path)
        board = Board.new #NUM_ROUNDS, NUM_CATEGORIES, NUM_QUESTIONS
        require 'parser'
        if board.load(Parser.new(@file))
          session[:board] = board
          session[:round] = 0
          session[:players] = @players
          flash[:notice] = "File loaded successfully."
          redirect_to :action => :board, :id => 'r1'
          return
        end
      rescue IOError => e
        flash[:notice] = e.message
      end
    end

    # Leave the image path empty if it is the same as the default path
    @image_path = '' if @image_path == default_image_path
  end

  ############################################################
  # PRIVATE
  ############################################################

  private

  def validate_session
    if !session[:board]
      redirect_to :action => :setup
      return
    end
    @board = session[:board]
    @round = session[:round] || 0
    @is_final = @board[@round].num_cells_total == 1
    @image_path = session[:image_path]
    @default_image_path = default_image_path
    @players = session[:players]
    set_image_path(@image_path)
  end

  def set_image_path(image_path)
    @image_path = image_path
    session[:image_path] = @image_path
    FormattedText.image_dir = @image_path
  end

  def increment_score(player_num, increment)
    @players[player_num].score += increment
  end

  def default_image_path
    path = render_to_string :inline => '<%= image_path("rails.png") -%>'
    path = path.strip.sub(/rails\.png.*$/, '')
    path
  end

  def score_value(id)
    (@round+1) * (id / @board.num_categories_per_round + 1) * 100
  end

  ############################################################
  # Request link handling
  ############################################################
  
  def GameController.req(rq, id = nil)
    "#{'%c' % rq}#{id ? id : ''}"
  end

  ##############################

  def req(rq, id = nil)
    GameController.req(rq, id)
  end

  ##############################
  
  def parse_req
    rq = params[:id]
    id = nil
    if rq
      id = rq[1..999].to_i
      rq = rq[0]
    end

    [rq, id]
  end

  ##############################

  def parse_selected_player
    player = params[:h_selected_player]
    raise GameException, 'Player unspecified' if !player
    player = player.to_i
    raise GameException, 'Player incorrectly specified' if player < 1 or player > NUM_PLAYERS
    player -= 1
    player
  end

  ##############################

  def parse_bids
    bids = params[:h_bids]
    raise GameException, 'Bid(s) unspecified' if !bids
    raise GameException, 'Bid(s) incorrectly specified' if (!bids.kind_of? Hash or bids.size != NUM_PLAYERS)
    raise GameException, 'Bid(s) must be non-negative' if (bids.find { |bid| bid[1].to_i < 0 })
    bids = (0...NUM_PLAYERS).collect { |i| bids[i.to_s].to_i }
    bids
  end

  ##############################

  def parse_choices
    choices = params[:h_choices]
    raise GameException, 'Choice(s) unspecified' if !choices
    raise GameException, 'Choice(s) incorrectly specified' if (!choices.kind_of? Hash or choices.size != NUM_PLAYERS)
    choices = (0...NUM_PLAYERS).collect { |i| choices[i.to_s].upcase }
    choices
  end

  ##############################

  public

  # Links to an image in the application directory given the file name and optional arguments, e.g. :class, :alt, :title, :style
  def image(filename, options = {})
    html = '<img'
    for key in options.keys
      html << " #{key.to_s}='#{options[key]}'" if options[key]
    end
    html << " src='#{@default_image_path}#{filename}' />"
  end

  private

  # Links to one of the feedback pictures for correct/wrong answers.
  def icon(filename)
    image("icons/#{filename}", :class => 'icon')
  end

  def random_icon(type)
    type = type.to_i
    icons = ICONS[type]
    return '' if !icons
    icon(icons[rand(icons.size)])
#    icon(icons[0])
  end

  ############################################################
  # HELPERS for the board action
  ############################################################

  # Set up the board grid for rendering
  def board_setup_for_rendering
    # Get the array of things to render
    round = @board[@round]

    # If there is only one cell in the round, then display only the category
    # with a link to the question.
    if round.num_cells_total == 1
      @cells = [[{ :value => round[0].title.html_text, :id => req(REQ_QUESTION, 0), :done => round[0][0].done  }]]
      
    # Otherwise, put categories at the top and a score associated with every cell
    else
      @cells = [ round.collect { |category| { :value => category.title.html_text, :id => nil, :done => nil } } ]
      for c in 0...round.num_cells_per_category do
        @cells << (0...round.num_categories).collect do |t|
          { :value => "#{(@round+1)*(c+1)}00",
            :id => req(REQ_QUESTION, c*round.num_categories+t),
            :done => round[t][c].done }
        end
      end
    end
  end

  ##############################

  # Set up the rounds for rendering
  def board_change_round(round)
    @round = [[round-1,0].max, @board.num_rounds-1].min
    session[:round] = @round
    @popup_link = url_for(:action => :board, :id => nil)
    @popup = render_to_string(:partial => 'game/board/popup/intro', :locals => { :intro => @board[@round].intro.html_text })
  end

  ##############################

  # Show the results of the game
  def board_done
    @winning_score = (@players.collect { |player| player.score }).max
    @winners = @players.select { |player| player.score == @winning_score }
    @winners = @winners.collect { |player| player.name }
    @popup = render_to_string(:partial => 'game/board/popup/done')
  end

  ##############################
  
  # Check if daily-double needs to be addressed. If not, set the new request
  # to showing the question
  def board_questions_and_answers(rq, id)
    if (id >= 0 and id < @board.num_categories_per_round * @board.num_cells_per_category)
      cat = @board[@round][id % @board.num_categories_per_round]
      cell = cat[id / @board.num_categories_per_round]
      cell.done = true
      category = cell.wild ? '<span style="font-size:15px"><h1>WILDCARD!!!</h1></span>' : cat.title.html_text + '<br />'

      ##################################################
      # FINAL
      if rq == REQ_QUESTION and @is_final
        valid = false

        # Here, h_steps is used to figure out how far in the final question
        # setup process we are.
        @steps = params[:h_steps].to_i

        if request.post?
          begin
            @bids = parse_bids
            for i in 0...@players.size
              bid = @bids[i]
              max_bid = @players[i].max_bet_amount
              raise GameException, 'Enter a non-negative bid amount' if bid < 0
              raise GameException, "#{@players[i].name} cannot bid more than #{max_bid}" if bid > max_bid
            end
            valid = true
            if @steps == 0
                flash[:notice] = '<b>Bids:</b> &nbsp; ' + @bids.join(' &nbsp; ') + '<br /><i>Close this message box!</i>'
            end
          rescue GameException => e
            valid = false
            flash[:notice] = e.message
          end
        end

        # If the information has not been successfully parsed, ask for it again
        if !valid or @steps == 0
          @hidden_form = {
            :action => :board,
            :link => req(REQ_QUESTION, id)
          }
          @steps = !valid ? 0 : @steps + 1
          @popup = render_to_string(:partial => 'game/board/popup/all_bet',
                                     :locals => { :category => category, :bids_valid => valid })
          rq = nil   # Make sure the question is not output
        end

      ##################################################
      # DAILY DOUBLE
      elsif rq == REQ_QUESTION and cell.daily_double
        valid = false

        if request.post?
          begin
            @selected_player = parse_selected_player  # Ensure that the player has been specified
            @bids = parse_bids
            bid = @bids[@selected_player]
            max_bid = @players[@selected_player].max_bet_amount
            raise GameException, 'Enter a non-negative bid amount' if bid < 0
            raise GameException, "Cannot bid more than #{max_bid}" if bid > max_bid
            valid = true
          rescue GameException => e
            valid = false
            flash[:notice] = e.message
          end
        end
 
        # If the information has not been successfully parsed, ask for it again
        if !valid
          @hidden_form = {
            :action => :board,
            :link => req(REQ_QUESTION, id),
            :selected_player => @selected_player
          }
          @popup = render_to_string(:partial => 'game/board/popup/daily_double',
                                     :locals => { :category => category })
          rq = nil   # Make sure the question is not output
        end
      end

      ##################################################
      # If a player GUESS has been made, evaluate it and update the score
      if rq == REQ_GUESS
        begin
          @bids = parse_bids
          @choices = parse_choices

          ##############################
          # FINAL
          if @is_final # Assuming multiple choice
            #flash[:notice] = @choices.inspect
            for player in 0...@players.size
              @steps = params[:h_steps].to_i

              choice = @choices[player][0]
              if choice == cell.ans
                increment_score(player, @bids[player])
              else
                increment_score(player, -@bids[player])
              end
              rq = REQ_ANSWER  # Show the answer
            end

          ##############################
          # REGULAR QUESTION
          else
            @selected_player = parse_selected_player  # Ensure that the player has been specified

            @steps = params[:h_steps].to_i

            increment = cell.daily_double ? @bids[@selected_player] : score_value(id)
            #decrement = (cell.daily_double or @steps > 0 or cell.binary? or !cell.multiple_choice?) ? -increment : -increment/2
            decrement = -increment

            choice = @choices[@selected_player][0]

            @steps += 1
            if cell.correct_answer?(choice)
              flash[:notice] = "<span class='pos_feedback'>Correct!</span><br />#{random_icon(ICON_POS)}"
              increment_score(@selected_player, increment)
              rq = REQ_ANSWER  # Show the answer
            else
              flash[:notice] = "<span class='neg_feedback'>WRONG!</span><br />#{random_icon(ICON_NEG)}"
              increment_score(@selected_player, decrement)
              rq = (@steps >= cell.max_num_guesses) ? REQ_ANSWER : REQ_QUESTION
              @selected_player = nil  # Don't highlight player so that another has to be chosen
            end
          end

        rescue GameException => e
          flash[:notice] = e.message
        end
      end

      ##################################################
      # Show the question or the answer
      if rq == REQ_QUESTION or rq == REQ_ANSWER
        @is_question = (rq == REQ_QUESTION)
        @is_answer   = (rq == REQ_ANSWER) # Used in _qna.rhtml

        if cell.daily_double and @selected_player
          category << @bids[@selected_player].to_s
        elsif !@is_final and !cell.daily_double
          category << score_value(id).to_s
        end

        @qna = {
          :title      => (@is_question ? 'Question' : 'Answer'),
          :category   => category,
          :link_prev  => (@is_question ? '' : req(REQ_QUESTION, id)),
          :link_next  => (@is_question ? req(REQ_ANSWER, id) : ''),
          :link_prev_title => (@is_question ? 'Close' : 'Question'),
          :link_next_title => (@is_question ? 'Answer' : 'Close'),
          :content    => (@is_question ? cell.question : cell.answer).html_text,
          :choice_buttons => (@is_question ? cell.choices(self, @is_final) : []),
          :correct_choice => (cell.ans ? ('%c' % cell.ans) : 'correct'),
        }
        @hidden_form = {
          :action => :board,
          :link => req(REQ_GUESS, id),
          :selected_player => @selected_player,
        }
        @popup = render_to_string(:partial => 'game/board/popup/qna')
      end
    end    
  end

  ############################################################
end
